package eu.ztsh.wymiana.web.controller;

import eu.ztsh.wymiana.exception.UserAlreadyExistsException;
import eu.ztsh.wymiana.exception.UserNotFoundException;
import eu.ztsh.wymiana.service.UserService;
import eu.ztsh.wymiana.util.UserMapper;
import eu.ztsh.wymiana.validation.ValidationFailedException;
import eu.ztsh.wymiana.web.model.UserCreateRequest;
import eu.ztsh.wymiana.web.model.UserResponse;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.media.Content;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.responses.ApiResponses;
import io.swagger.v3.oas.annotations.tags.Tag;
import jakarta.validation.Valid;
import jakarta.validation.ValidationException;
import lombok.RequiredArgsConstructor;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.validation.annotation.Validated;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

@RequiredArgsConstructor
@Validated
@RestController
@RequestMapping(path = "/api/user", produces = "application/json")
@Tag(name="User management", description = "Create or get user")
public class UserController {

    private final UserService userService;

    @Operation(summary = "Get user by PESEL")
    @ApiResponses(value = {
        @ApiResponse(responseCode = "200", description = "User found"),
        @ApiResponse(responseCode = "400", description = "Request not valid", content = @Content),
        @ApiResponse(responseCode = "404", description = "User not found", content = @Content),
        @ApiResponse(responseCode = "500", description = "Another error has occurred", content = @Content)
    })
    @GetMapping("{pesel}")
    public ResponseEntity<UserResponse> get(@PathVariable("pesel") String pesel) {
        try {
            return userService.get(pesel)
                .map(UserMapper::pojoToResponse)
                .map(ResponseEntity::ok)
                .orElseThrow(() -> new UserNotFoundException(pesel));
        } catch (Exception e) {
            return ResponseEntity.status(switch (e) {
                case ValidationException ignore -> 400;
                case UserNotFoundException ignore -> 404;
                default -> 500;
            }).build();
        }
    }

    @Operation(summary = "Create user")
    @ApiResponses(value = {
        @ApiResponse(responseCode = "204", description = "User created", content = @Content),
        @ApiResponse(responseCode = "400", description = "Request not valid", content = @Content(mediaType = MediaType.TEXT_PLAIN_VALUE)),
        @ApiResponse(responseCode = "409", description = "User already exists", content = @Content(mediaType = MediaType.TEXT_PLAIN_VALUE)),
        @ApiResponse(responseCode = "500", description = "Another error has occurred", content = @Content(mediaType = MediaType.TEXT_PLAIN_VALUE))
    })
    @PostMapping
    public ResponseEntity<String> create(@Valid @RequestBody UserCreateRequest request) {
        try {
            userService.create(request);
        } catch (Exception e) {
            var status = switch (e) {
                case ValidationFailedException ignored -> HttpStatus.BAD_REQUEST;
                case UserAlreadyExistsException ignored -> HttpStatus.CONFLICT;
                default -> HttpStatus.INTERNAL_SERVER_ERROR;
            };
            return ResponseEntity.status(status).body(e.getMessage());
        }
        return ResponseEntity.status(204).build();
    }

}
