package eu.ztsh.wymiana;

import eu.ztsh.wymiana.data.entity.CurrencyEntity;
import eu.ztsh.wymiana.data.entity.UserEntity;
import eu.ztsh.wymiana.model.Currency;
import eu.ztsh.wymiana.model.Rate;
import eu.ztsh.wymiana.model.Rates;
import eu.ztsh.wymiana.model.User;
import eu.ztsh.wymiana.web.model.CurrencyExchangeRequest;
import eu.ztsh.wymiana.web.model.UserCreateRequest;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

public class EntityCreator {

    public static class Constants {

        public static String PESEL = "00281018264";
        public static String INVALID_PESEL = PESEL.replace('6', '7');
        public static String ANOTHER_PESEL = "08280959342";
        public static String NAME = "Janina";
        public static String SURNAME = "Kowalska";
        public static BigDecimal PLN = createScaled(20.10, 2);
        public static BigDecimal USD_SELL = createScaled(5.18, 2);
        public static BigDecimal USD_BUY = createScaled(5.08, 2);
        public static String PLN_SYMBOL = "PLN";
        public static String USD_SYMBOL = "USD";
        public static BigDecimal BUY_RATE = createScaled(3.8804, 4);
        public static BigDecimal SELL_RATE = createScaled(3.9572, 4);

    }

    public static UserEntityBuilder userEntity() {
        return new UserEntityBuilder();
    }

    public static User user() {
        return user(Constants.PLN, BigDecimal.ZERO);
    }

    public static User user(BigDecimal pln, BigDecimal usd) {
        Map<String, Currency> currencies = new HashMap<>();
        if (pln.signum() > 0) {
            currencies.put("PLN", new Currency("PLN", pln));
        }
        if (usd.signum() > 0) {
            currencies.put("USD", new Currency("USD", usd));
        }
        return new User(Constants.NAME, Constants.SURNAME, Constants.PESEL, currencies);
    }

    public static UserCreateRequest.UserCreateRequestBuilder userRequest() {
        return userRequest(null);
    }

    public static UserCreateRequest.UserCreateRequestBuilder userRequest(Double initial) {
        return UserCreateRequest.builder().name(Constants.NAME)
            .surname(Constants.SURNAME)
            .pesel(Constants.PESEL)
            .initial(initial != null ? createScaled(initial, 2) : Constants.PLN);
    }

    public static CurrencyExchangeRequest.CurrencyExchangeRequestBuilder exchangeRequest() {
        return CurrencyExchangeRequest.builder().pesel(Constants.PESEL);
    }

    public static Rates rates(String date) {
        var rates = new Rates();
        rates.setTable("C");
        rates.setCurrency("dolar amerykański");
        rates.setCode("USD");
        var rate = new Rate();
        rate.setNo("096/C/NBP/2024");
        rate.setEffectiveDate(date);
        rate.setBid(Constants.BUY_RATE);
        rate.setAsk(Constants.SELL_RATE);
        rates.setRates(List.of(rate));
        return rates;
    }

    public static class UserEntityBuilder {

        String name;
        String surname;
        String pesel;
        BigDecimal pln = BigDecimal.valueOf(-1);
        BigDecimal usd = BigDecimal.valueOf(-1);

        public UserEntityBuilder name(String name) {
            this.name = name;
            return this;
        }

        public UserEntityBuilder surname(String surname) {
            this.surname = surname;
            return this;
        }

        public UserEntityBuilder pesel(String pesel) {
            this.pesel = pesel;
            return this;
        }

        public UserEntityBuilder pln(double pln) {
            return this.pln(BigDecimal.valueOf(pln));
        }

        public UserEntityBuilder pln(BigDecimal pln) {
            this.pln = pln.setScale(2, RoundingMode.HALF_UP);
            return this;
        }

        public UserEntityBuilder usd(double usd) {
            return this.usd(BigDecimal.valueOf(usd));
        }

        public UserEntityBuilder usd(BigDecimal usd) {
            this.usd = usd.setScale(2, RoundingMode.HALF_UP);
            return this;
        }

        public UserEntity build() {
            var nonnulPesel = Optional.ofNullable(pesel).orElse(Constants.PESEL);
            List<CurrencyEntity> currencies = new ArrayList<>();
            if (pln.signum() > -1) {
                currencies.add(new CurrencyEntity(nonnulPesel, "PLN", pln));
            }
            if (usd.signum() > -1) {
                currencies.add(new CurrencyEntity(nonnulPesel, "USD", usd));
            }
            if (currencies.isEmpty()) {
                currencies.add(new CurrencyEntity(nonnulPesel, "PLN", Constants.PLN));
            }
            return new UserEntity(
                nonnulPesel,
                Optional.ofNullable(name).orElse(Constants.NAME),
                Optional.ofNullable(surname).orElse(Constants.SURNAME),
                currencies
            );
        }

    }

    private static BigDecimal createScaled(double input, int scale) {
        return BigDecimal.valueOf(input).setScale(scale, RoundingMode.HALF_UP);
    }

}
